//
// Created by bytedance on 2021/9/28.
//
#include <string>
#include <vector>
#include <iostream>
#define DR_WAV_IMPLEMENTATION
#include "dr_libs/dr_wav.h"
#include "sami_core.h"
#include "help_function.h"

#define BLOCK_SIZE 512

using namespace std;

string noise_fun(string func_id,string input_file,string output_file,string res_path){
    string res_str = "OK";
    cout << "func_id:" << func_id << endl;
    cout << "input_file:" << input_file << endl;
    cout << "output_file:" << output_file << endl;
    cout << "res_path:" << res_path << endl;

    //1. 输入输出文件准备 ===================================
    string in_wav_path = input_file;
    string out_wav_path = output_file;
    drwav in_wav;
    drwav_bool32 drwav_ret = drwav_init_file(&in_wav, in_wav_path.c_str(), NULL);
    if(drwav_ret == false){
        cout << "open input_file error!!!" << endl;
        res_str = "open input_file error";
        return res_str;
    }

    cout << "===input file info====" << endl;
    cout << "ch:" << in_wav.channels <<endl;
    cout << "sample rate:" << in_wav.sampleRate << endl;
    cout << "bitsPerSample:" << in_wav.bitsPerSample << endl;
    cout << "translatedFormatTag:" << in_wav.translatedFormatTag << endl;
    cout << "===input file info====" << endl;

    drwav out_wav;
    drwav_data_format format;
    format.container = drwav_container_riff;
    format.format = DR_WAVE_FORMAT_PCM;
    format.channels = in_wav.channels;
    format.sampleRate = in_wav.sampleRate;
    format.bitsPerSample = 16;
    drwav_ret = drwav_init_file_write(&out_wav, out_wav_path.c_str(), &format, NULL);
    if(drwav_ret == false){
        cout << "open output_file error!!!" << endl;
        drwav_uninit(&in_wav);
        res_str = "open input_file error";
        return res_str;
    }

    int num_frame = BLOCK_SIZE;
    int f32_buf_num = num_frame * in_wav.channels;
    vector<float> f32_buffer;
    f32_buffer.resize(f32_buf_num);
    vector<signed short> f32_to_s16_buffer;
    f32_to_s16_buffer.resize(f32_buf_num*(sizeof(float )/sizeof(signed short)));

    //2. load model and init handle ===================================
    std::string model_path;
    identify identify;
    if (func_id == "DENOISE_V2_SPEECH44k"){
        if (in_wav.sampleRate==44100){
            model_path = res_path + "model/denoise_v2/unet_denoise_44k_speech_model_v1.0.model";
            identify = SAMICoreIdentify::SAMICoreIdentify_DENOISE_V2_SPEECH44k;
        }else{
            cout << "nonsupport sample rate" << endl;
            res_str = "nonsupport sample rate";
        }
    }else if(func_id == "DENOISE_V2_SPEECH16k"){
        if(in_wav.sampleRate==16000){
            model_path = res_path + "model/denoise_v2/unet_denoise_16k_speech_model_v1.0.model";
            identify = SAMICoreIdentify::SAMICoreIdentify_DENOISE_V2_SPEECH16k;
        }else{
            cout << "nonsupport sample rate" << endl;
            res_str = "nonsupport sample rate";
        }
    }else if (func_id == "DENOISE_V2_MUSIC44k"){
        if (in_wav.sampleRate==44100){
            model_path = res_path + "model/denoise_v2/unet_denoise_44k_music_model_v1.0.model";
            identify = SAMICoreIdentify::SAMICoreIdentify_DENOISE_V2_MUSIC44k;
        }else{
            cout << "nonsupport sample rate" << endl;
            res_str = "nonsupport sample rate";
        }
    }else if(func_id == "DENOISE_V2_MUSIC16k"){
        if(in_wav.sampleRate==16000){
            model_path = res_path + "model/denoise_v2/unet_denoise_16k_music_model_v1.0.model";
            identify = SAMICoreIdentify::SAMICoreIdentify_DENOISE_V2_MUSIC16k;
        }else{
            cout << "nonsupport sample rate" << endl;
            res_str = "nonsupport sample rate";
        }
    }else if(func_id == "DENOISE_V2_DEREVERB44k"){
        if(in_wav.sampleRate==44100){
            model_path = res_path + "model/denoise_v2/unet_denoise_44k_dereverb_model_v1.0.model";
            identify = SAMICoreIdentify::SAMICoreIdentify_DENOISE_V2_DEREVERB44k;
        }else{
            cout << "nonsupport sample rate" << endl;
            res_str = "nonsupport sample rate";
        }
    }else{
        cout << "nonsupport func_id" << endl;
        res_str = "nonsupport sample rate";
    }

    if (res_str!="OK"){
        return res_str;
    }

    std::vector<uint8_t> model_buf = loadModelAsBinary(model_path);
    if (model_buf.size() <= 0){
        cout << "open model file error:" << model_path << endl;
        res_str = "open model file error";
        return res_str;
    }else{
        cout << "open model success,module size:" << model_buf.size() << endl;
    }

    const int block_size = BLOCK_SIZE;
    const int sample_rate = in_wav.sampleRate;
    const int num_channels = in_wav.channels;
    SAMICoreHandle handle = nullptr;
    executorContextCreateParameter createParameter;
    createParameter.sampleRate = sample_rate;
    createParameter.maxBlockSize = block_size;
    createParameter.numChannel = num_channels;
    createParameter.modelBuffer = reinterpret_cast<char*>(model_buf.data());
    createParameter.modelLen = model_buf.size();
    int ret = SAMICoreCreateHandleByIdentify(&handle,identify, &createParameter);
    if (ret != SAMI_OK){
        cout << "create handle fail,ret:" << ret << endl;
        res_str = "create handle fail:"+ to_string(ret);
        return "res_str";
    }else{
        cout << "create handle success" << endl;
    }

    //3. init process buffer ===================================
    SAMICoreAudioBuffer in_audio_buffer;
    in_audio_buffer.numberChannels = num_channels;
    in_audio_buffer.numberSamples = block_size;
    in_audio_buffer.data = new float*[num_channels];
    in_audio_buffer.isInterleave = 0;

    SAMICoreAudioBuffer out_audio_buffer;
    out_audio_buffer.numberChannels = num_channels;
    out_audio_buffer.numberSamples = block_size;
    out_audio_buffer.data = new float*[num_channels];
    out_audio_buffer.isInterleave = 0;

    for(int c = 0; c < int(num_channels); ++c) {
        in_audio_buffer.data[c] = new float[block_size];
        out_audio_buffer.data[c] = new float[block_size];
    }

    SAMICoreBlock in_block;
    memset(&in_block, 0, sizeof(SAMICoreBlock));
    in_block.numberAudioData = 1;
    in_block.dataType = SAMICoreDataType::SAMICoreDataType_AudioBuffer;
    in_block.audioData = &in_audio_buffer;

    SAMICoreBlock out_block;
    memset(&out_block, 0, sizeof(SAMICoreBlock));
    out_block.numberAudioData = 1;
    out_block.dataType = SAMICoreDataType::SAMICoreDataType_AudioBuffer;
    out_block.audioData = &out_audio_buffer;

    //4. process  ===================================
    cout << "process start" << endl;
    int readed_frames = 0;
    do {
        //read from file
        readed_frames = drwav_read_pcm_frames_f32(&in_wav, num_frame, f32_buffer.data());
        if (readed_frames<=0){
            break;
        }
        //Interleave to Planar
        interleaveToPlanarFloat(f32_buffer.data(),in_audio_buffer.data, readed_frames, num_channels);

        //process
        int ret = SAMICoreProcess(handle, &in_block, &out_block);
        if (ret!=SAMI_OK){
            res_str = "SAMICoreProcess error:" + to_string(ret) ;
            cout << res_str << endl;
            break;
        }

        //Planar to Interleave
        planarToInterleaveFloat(const_cast<const float **>(out_audio_buffer.data),f32_buffer.data(),readed_frames,num_channels);

        //f32->s16
        drwav_f32_to_s16(f32_to_s16_buffer.data(), f32_buffer.data(), readed_frames*num_channels);

        //write to out file
        drwav_uint64 writed_frames = drwav_write_pcm_frames(&out_wav, readed_frames, f32_to_s16_buffer.data());

    }while (readed_frames!=0);
    cout << "process end" << endl;

    //5. release
    cout << "release" << endl;
    drwav_uninit(&in_wav);
    drwav_uninit(&out_wav);

    ret = SAMICoreDestroyHandle(handle);
    if(ret!=SAMI_OK){
        res_str = "SAMICoreDestroyHandle error:" + to_string(ret) ;
        cout << res_str << endl;
    }

    for(int c = 0; c < int(num_channels); ++c) {
        delete[] in_audio_buffer.data[c];
        delete[] out_audio_buffer.data[c];
    }

    delete[] in_audio_buffer.data;
    delete[] out_audio_buffer.data;

    return res_str;
}

